@@ -205,7 +205,7 @@ fn callback(state_ptr: *const InnerCell<CallFutureState<Vec<u8>>>) {
     // Make sure to un-borrow_mut the state.
     {
         state.borrow_mut().result = Some(match reject_code() {
-            RejectionCode::NoError => unsafe { Ok(arg_data_raw()) },
+            RejectionCode::NoError => Ok(arg_data_raw()),
             n => Err((n, reject_message())),
         });
     }
@@ -222,23 +222,109 @@ fn callback(state_ptr: *const InnerCell<CallFutureState<Vec<u8>>>) {
 /// Waker is a very opaque API, so the best we can do is set a global flag and proceed normally.
 fn cleanup(state_ptr: *const InnerCell<CallFutureState<Vec<u8>>>) {
     let state = unsafe { WasmCell::from_raw(state_ptr) };
-    // We set the call result, even though it won't be read on the default executor, because we can't guarantee it was called on our executor.
-    // None of these calls trap - the rollback from the previous trap ensures that the Mutex is not in a poisoned state.
-    {
-        state.borrow_mut().result = Some(match reject_code() {
-            RejectionCode::NoError => unsafe { Ok(arg_data_raw()) },
-            n => Err((n, reject_message())),
-        });
-    }
+    // We set the call result, even though it won't be read on the
+    // default executor, because we can't guarantee it was called on
+    // our executor. However, we are not allowed to inspect
+    // reject_code() inside of a cleanup callback, so always set the
+    // result to a reject.
+    //
+    // Borrowing does not trap - the rollback from the
+    // previous trap ensures that the WasmCell can be borrowed again.
+    state.borrow_mut().result = Some(Err((RejectionCode::NoError, "cleanup".to_string())));
     let w = state.borrow_mut().waker.take();
     if let Some(waker) = w {
-        // Flag that we do not want to actually wake the task - we want to drop it *without* executing it.
+        // Flag that we do not want to actually wake the task - we
+        // want to drop it *without* executing it.
         crate::futures::CLEANUP.store(true, Ordering::Relaxed);
         waker.wake();
         crate::futures::CLEANUP.store(false, Ordering::Relaxed);
     }
 }
 
+fn add_payment(payment: u128) {
+    if payment == 0 {
+        return;
+    }
+    let high = (payment >> 64) as u64;
+    let low = (payment & u64::MAX as u128) as u64;
+    unsafe {
+        ic0::call_cycles_add128(high as i64, low as i64);
+    }
+}
+
+/// Sends a one-way message with `payment` cycles attached to it that invokes `method` with
+/// arguments `args` on the principal identified by `id`, ignoring the reply.
+///
+/// Returns `Ok(())` if the message was successfully enqueued, otherwise returns a reject code.
+///
+/// # Notes
+///
+///   * The caller has no way of checking whether the destination processed the notification.
+///     The system can drop the notification if the destination does not have resources to
+///     process the message (for example, if it's out of cycles or queue slots).
+///
+///   * The callee cannot tell whether the call is one-way or not.
+///     The callee must produce replies for all incoming messages.
+///
+///   * It is safe to upgrade a canister without stopping it first if it sends out *only*
+///     one-way messages.
+///
+///   * If the payment is non-zero and the system fails to deliver the notification, the behaviour
+///     is unspecified: the funds can be either reimbursed or consumed irrevocably by the IC depending
+///     on the underlying implementation of one-way calls.
+pub fn notify_with_payment128<T: ArgumentEncoder>(
+    id: Principal,
+    method: &str,
+    args: T,
+    payment: u128,
+) -> Result<(), RejectionCode> {
+    let args_raw = encode_args(args).expect("failed to encode arguments");
+    notify_raw(id, method, &args_raw, payment)
+}
+
+/// Like [notify_with_payment128], but sets the payment to zero.
+pub fn notify<T: ArgumentEncoder>(
+    id: Principal,
+    method: &str,
+    args: T,
+) -> Result<(), RejectionCode> {
+    notify_with_payment128(id, method, args, 0)
+}
+
+/// Like [notify], but sends the argument as raw bytes, skipping Candid serialization.
+pub fn notify_raw(
+    id: Principal,
+    method: &str,
+    args_raw: &[u8],
+    payment: u128,
+) -> Result<(), RejectionCode> {
+    let callee = id.as_slice();
+    // We set all callbacks to -1, which is guaranteed to be invalid callback index.
+    // The system will still deliver the reply, but it will trap immediately because the callback
+    // is not a valid function. See
+    // https://www.joachim-breitner.de/blog/789-Zero-downtime_upgrades_of_Internet_Computer_canisters#one-way-calls
+    // for more context.
+    let err_code = unsafe {
+        ic0::call_new(
+            callee.as_ptr() as i32,
+            callee.len() as i32,
+            method.as_ptr() as i32,
+            method.len() as i32,
+            /* reply_fun = */ -1,
+            /* reply_env = */ -1,
+            /* reject_fun = */ -1,
+            /* reject_env = */ -1,
+        );
+        add_payment(payment);
+        ic0::call_data_append(args_raw.as_ptr() as i32, args_raw.len() as i32);
+        ic0::call_perform()
+    };
+    match err_code {
+        0 => Ok(()),
+        c => Err(RejectionCode::from(c)),
+    }
+}
+
 /// Similar to `call`, but without serialization.
 pub fn call_raw(
     id: Principal,
@@ -263,13 +349,7 @@ pub fn call_raw128(
     payment: u128,
 ) -> impl Future<Output = CallResult<Vec<u8>>> {
     call_raw_internal(id, method, args_raw, move || {
-        if payment > 0 {
-            unsafe {
-                let high = (payment >> 64) as u64;
-                let low = (payment & u64::MAX as u128) as u64;
-                ic0::call_cycles_add128(high as i64, low as i64);
-            }
-        }
+        add_payment(payment);
     })
 }
 
@@ -364,8 +444,9 @@ pub fn call_with_payment128<T: ArgumentEncoder, R: for<'a> ArgumentDecoder<'a>>(
 /// and [reject_message()] if it failed.
 pub fn result<T: for<'a> ArgumentDecoder<'a>>() -> Result<T, String> {
     match reject_code() {
-        RejectionCode::NoError => decode_args(&unsafe { arg_data_raw() })
-            .map_err(|e| format!("Failed to decode arguments: {}", e)),
+        RejectionCode::NoError => {
+            decode_args(&arg_data_raw()).map_err(|e| format!("Failed to decode arguments: {}", e))
+        }
         _ => Err(reject_message()),
     }
 }
@@ -474,16 +555,27 @@ pub fn msg_cycles_accept128(max_amount: u128) -> u128 {
 }
 
 /// Returns the argument data as bytes.
-pub(crate) unsafe fn arg_data_raw() -> Vec<u8> {
-    let len: usize = ic0::msg_arg_data_size() as usize;
-    let mut bytes = vec![0u8; len as usize];
-    ic0::msg_arg_data_copy(bytes.as_mut_ptr() as i32, 0, len as i32);
-    bytes
+pub fn arg_data_raw() -> Vec<u8> {
+    unsafe {
+        let len: usize = ic0::msg_arg_data_size() as usize;
+        let mut bytes = Vec::with_capacity(len);
+        ic0::msg_arg_data_copy(bytes.as_mut_ptr() as i32, 0, len as i32);
+        bytes.set_len(len);
+        bytes
+    }
+}
+
+/// Replies with the bytes passed
+pub fn reply_raw(buf: &[u8]) {
+    unsafe {
+        ic0::msg_reply_data_append(buf.as_ptr() as i32, buf.len() as i32);
+        ic0::msg_reply();
+    }
 }
 
 /// Returns the argument data in the current call.
 pub fn arg_data<R: for<'a> ArgumentDecoder<'a>>() -> R {
-    let bytes = unsafe { arg_data_raw() };
+    let bytes = arg_data_raw();
 
     match decode_args(&bytes) {
         Err(e) => trap(&format!("{:?}", e)),
